#include <iostream>
#include <algorithm>
#include <vector>
#include <memory>
#include <cassert>

struct Etat
{
   virtual void entrer() = 0;
   virtual void dedans() = 0;
   virtual void sortir() = 0;
   virtual ~Etat() throw()
      { }
};

class Pas
{
   int nb_;
public:
   Pas()
      : nb_(0)
   {
   }
   int nb_pas() const
      { return nb_; }
   void marcher()
      { ++nb_; }
};


class EtatMouvement
   : public Etat
{
   Pas &pas_;
public:
   EtatMouvement(Pas &p)
      : pas_(p)
   {
   }
   int nb_pas() const
      { return pas_.nb_pas(); }
   void dedans()
      { pas_.marcher(); }
};

class VaAGauche
   : public EtatMouvement
{
public:
   VaAGauche(Pas &p)
      : EtatMouvement(p)
   {
   }
   void entrer()
      { std::cout << "Tourner vers la gauche" << std::endl; }
   void dedans()
   {
      EtatMouvement::dedans();
      std::cout << "Aller vers la gauche" << std::endl;
   }
   void sortir()
      { std::cout << "Arreter d'aller vers la gauche" << std::endl; }
};
class VaADroite
   : public EtatMouvement
{
public:
   VaADroite(Pas &p)
      : EtatMouvement(p)
   {
   }
   void entrer()
      { std::cout << "Tourner vers la droite" << std::endl; }
   void dedans()
   {
      EtatMouvement::dedans();
      std::cout << "Aller vers la droite" << std::endl;
   }
   void sortir()
      { std::cout << "Arreter d'aller vers la droite" << std::endl; }
};


struct Predicat
{
   virtual bool verifier(const std::shared_ptr<Etat>) const = 0;
   virtual ~Predicat() throw()
      { }
};

class Regle;
class EvaluerRegle;

class Transition
{
   std::shared_ptr<Etat> avant_;
   std::vector<std::unique_ptr<Regle>> regles_;
public:
   Transition(std::shared_ptr<Etat> avant)
      : avant_(avant)
   {
   }
   void ajouter_regle(std::unique_ptr<Regle> && r)
   {
      if (r) regles_.push_back(std::move(r));
   }
   bool a_prochaine() const;
   std::shared_ptr<Etat> courant()
      { return avant_; }
   std::shared_ptr<Transition> prochaine();
};


class Regle
{
   std::unique_ptr<Predicat> pred_;
   std::shared_ptr<Transition> prochaine_;
public:
   Regle(std::unique_ptr<Predicat> && pred, std::shared_ptr<Transition> prochaine = 0)
      : pred_(std::move(pred)), prochaine_(prochaine)
   {
      assert(pred_);
   }
   bool operator()(std::shared_ptr<Etat> e) const
      { return pred_->verifier(e); }
   std::shared_ptr<Transition> prochaine()
      { return prochaine_; }
};


bool Transition::a_prochaine() const
{
   using std::unique_ptr;
   using std::begin;
   using std::end;
   using std::find_if;
   return find_if(begin(regles_), end(regles_), [&](const unique_ptr<Regle> &r) {
      return (*r)(avant_);
   }) != end(regles_);
}


std::shared_ptr<Transition> Transition::prochaine()
{
   using std::unique_ptr;
   using std::begin;
   using std::end;
   using std::find_if;
   auto itt = find_if(begin(regles_), end(regles_), [&](const unique_ptr<Regle> &r) {
                 return (*r)(avant_);
              });
   return itt == end(regles_)? 0 : (*itt)->prochaine();
}



class Automate
{
   std::vector<std::shared_ptr<Etat>> etats_;
   std::vector<std::shared_ptr<Transition>> transitions_;
   std::shared_ptr<Transition> cur_;
public:
   Automate& ajouter(std::shared_ptr<Etat> e)
   {
      if (e) etats_.push_back (e);
      return *this;
   }
   Automate& ajouter(std::shared_ptr<Transition> t)
   {
      if (t) transitions_.push_back (t);
      return *this;
   }
   void executer()
   {
      if (!transitions_.empty())
      {
         cur_ = transitions_.front();
         cur_->courant ()->entrer();
         while(cur_)
         {
            cur_->courant()->dedans();
            if (cur_->a_prochaine())
            {
               cur_->courant()->sortir();
               cur_ = cur_->prochaine();
            }
         }
      }
   }
};


struct ChangementDirection
   : Predicat
{
   bool verifier(const std::shared_ptr<Etat> p) const
      { return static_cast<const EtatMouvement*>(p.get())->nb_pas() % 5 == 0; }
};


struct Essouflement
   : Predicat
{
   bool verifier(const std::shared_ptr<Etat> p) const
      { return static_cast<const EtatMouvement*>(p.get())->nb_pas() >= 30; }
};

template <class P>
   std::unique_ptr<Regle> creer_pred(P p)
   {
      using std::unique_ptr;
      return unique_ptr<Regle>(
         new Regle(unique_ptr<Predicat>(p))
      );
   }
template <class P>
   std::unique_ptr<Regle> creer_pred(P p, std::shared_ptr<Transition> t)
   {
      using std::unique_ptr;
      return unique_ptr<Regle>(
         new Regle(unique_ptr<Predicat>(p), t)
      );
   }
int main()
{
   using std::unique_ptr;
   using std::shared_ptr;
   Pas p;
   {
      Automate a;
      shared_ptr<Etat> p0 (new VaAGauche(p));
      shared_ptr<Etat> p1 (new VaADroite(p));
      shared_ptr<Transition> tG(new Transition(p0)),
                             tD(new Transition(p1));
      tG->ajouter_regle(creer_pred(new Essouflement));
      tG->ajouter_regle(creer_pred(new ChangementDirection, tD));
      tD->ajouter_regle(creer_pred(new Essouflement));
      tD->ajouter_regle(creer_pred(new ChangementDirection, tG));
      a.ajouter(p0).ajouter(p1).ajouter(tG).ajouter(tD);
      //
      // Go!
      //
      a.executer();
   }
} 